#include <fxos/vspace.h>
#include <fxos/os.h>
#include <cstring>

namespace FxOS {

//---
//  Simulated memory access primitives
//---

bool AbstractMemory::covers(MemoryRegion const &region) const noexcept
{
	return covers(region.start, region.size());
}

Addressable<int8_t> AbstractMemory::read_i8(uint32_t addr) const
{
	int8_t *i8 = (int8_t *)translate(addr, 1);
	if(!i8)
		return Addressable((int8_t)-1);
	return Addressable(addr, *i8);
}

Addressable<uint8_t> AbstractMemory::read_u8(uint32_t addr) const
{
	uint8_t *u8 = (uint8_t *)translate(addr, 1);
	if(!u8)
		return Addressable((uint8_t)-1);
	return Addressable(addr, *u8);
}

Addressable<int16_t> AbstractMemory::read_i16(uint32_t addr) const
{
	uint8_t *i16 = (uint8_t *)translate(addr, 2);
	if(!i16)
		return Addressable((int16_t)-1);
	int16_t v = (i16[0] << 8) | i16[1];
	return Addressable(addr, v);
}

Addressable<uint16_t> AbstractMemory::read_u16(uint32_t addr) const
{
	uint8_t *u16 = (uint8_t *)translate(addr, 2);
	if(!u16)
		return Addressable((uint16_t)-1);
	uint16_t v = (u16[0] << 8) | u16[1];
	return Addressable(addr, v);
}

Addressable<int32_t> AbstractMemory::read_i32(uint32_t addr) const
{
	uint8_t *i32 = (uint8_t *)translate(addr, 4);
	if(!i32)
		return Addressable((int32_t)-1);
	int32_t v = (i32[0] << 24) | (i32[1] << 16) | (i32[2] << 8) | i32[3];
	return Addressable(addr, v);
}

Addressable<uint32_t> AbstractMemory::read_u32(uint32_t addr) const
{
	uint8_t *u32 = (uint8_t *)translate(addr, 4);
	if(!u32)
		return Addressable((uint32_t)-1);
	uint32_t v = (u32[0] << 24) | (u32[1] << 16) | (u32[2] << 8) | u32[3];
	return Addressable(addr, v);
}

Addressable<std::string> AbstractMemory::read_str(uint32_t addr, size_t len)
	const
{
	char const *str = translate(addr, len);
	if(!str)
		return Addressable(std::string());
	return Addressable(addr, std::string(str, len));
}

//---
//  Bindings of data buffers into memory regions
//---

Binding::Binding(MemoryRegion source_region, Buffer source_buffer):
	region(source_region), buffer(source_buffer)
{
	/* Extend the buffer if it's not at least as large as the region */
	if(buffer.size < region.size()) {
		buffer = Buffer(buffer, region.size());
	}
}

bool Binding::covers(uint32_t addr, int size) const noexcept
{
	return size >= 0 && addr >= region.start && addr + size <= region.end;
}

bool Binding::covers(MemoryRegion const &r) const noexcept
{
	return covers(r.start, r.size());
}

char const *Binding::translate(uint32_t addr, int size) const
{
	if(!covers(addr, size)) return nullptr;
	return buffer.data.get() + (addr - region.start);
}

char const *Binding::translate_dynamic(uint32_t addr, int *size) const
{
	if(!covers(addr, 1)) return nullptr;
	*size = region.end - addr;
	return buffer.data.get() + (addr - region.start);
}

uint32_t Binding::search(uint32_t start, uint32_t end, void const *pattern,
	int size) const
{
	void const *data = translate(start, end-start);
	if(!data)
		return end;
	void const *occurrence = memmem(data, end - start, pattern, size);

	if(!occurrence)
		return end;
	return start + ((char *)occurrence - (char *)data);
}

//---
//  Composite memory targets
//---

VirtualSpace::VirtualSpace():
	m_bindings {}, m_buffers {}, m_os {nullptr}
{
}

VirtualSpace::VirtualSpace(Target const &target,
	std::vector<std::string> const &folders):
	VirtualSpace()
{
	for(auto binding: target.bindings)
	{
		MemoryRegion region = binding.first;
		ssize_t size = (region.size() > 0 ? region.size() : -1);

		auto b = m_buffers.emplace(m_buffers.end(),
			binding.second, folders, size);
		bind_region(region, *b);
	}

	this->mpu = target.mpu;
}

OS *VirtualSpace::os_analysis(bool force)
{
	if(!m_os || force) {
		m_os = std::make_unique<OS>(*this);
		/* We don't keep an OS analysis result that failed */
		if(m_os->type == OS::UNKNOWN)
			m_os = nullptr;
	}
	return m_os.get();
}

void VirtualSpace::bind_region(MemoryRegion const &region,Buffer const &buffer)
{
	Binding b(region, buffer);
	m_bindings.push_back(b);
}

bool VirtualSpace::covers(uint32_t addr, int size) const noexcept
{
	for(auto it = m_bindings.crbegin(); it != m_bindings.crend(); it++)
	{
		if(it->covers(addr, size)) return true;
	}

	return false;
}

bool VirtualSpace::covers(MemoryRegion const &region) const noexcept
{
	return covers(region.start, region.size());
}

char const *VirtualSpace::translate(uint32_t addr, int size) const
{
	for(auto it = m_bindings.crbegin(); it != m_bindings.crend(); it++)
	{
		char const *ptr = it->translate(addr, size);
		if(ptr) return ptr;
	}
	return nullptr;
}

char const *VirtualSpace::translate_dynamic(uint32_t addr, int *size) const
{
	for(auto it = m_bindings.crbegin(); it != m_bindings.crend(); it++)
	{
		char const *ptr = it->translate_dynamic(addr, size);
		if(ptr) return ptr;
	}
	return nullptr;
}

uint32_t VirtualSpace::search(uint32_t start, uint32_t end,
	void const *pattern, int size) const
{
	uint32_t occurrence;
	if(end < start || !covers(start, end - start))
		return end;

	for(auto it = m_bindings.crbegin(); it != m_bindings.crend(); it++)
	{
		if(it->covers(start, end - start))
		{
			occurrence = it->search(start, end, pattern, size);
			if(occurrence != end) return occurrence;
		}
	}

	return end;
}

} /* namespace FxOS */
